import numpy as np
cimport numpy as np
cimport cython
from libcpp.queue cimport priority_queue
from libcpp.vector cimport vector
from libcpp.pair cimport pair

from collections import deque
import heapq

ctypedef pair[double, vector[int]] PAIR

cdef np.ndarray create_neighbor_mask(int length=1):
    mesh_range = np.arange(-length, length+1)
    # mesh_range = mesh_range[np.argsort(np.abs(mesh_range))]
    xx, yy, aa = np.meshgrid(mesh_range, mesh_range, mesh_range)
    return np.vstack((xx.flatten(),yy.flatten(),aa.flatten())).T

cdef double convert_180(double rad360):
    '''
    [0,360] -> [-180,180]
    '''
    return ((rad360 - np.pi) % (np.pi*2)) - np.pi

cdef np.ndarray[np.int64_t] create_circum_mask(int r):
    cdef np.ndarray[np.int64_t, ndim=2] x, y
    x, y = np.meshgrid(np.arange(-r,r+1), np.arange(-r,r+1))
    cdef np.ndarray[np.uint8_t, ndim=2] mask = np.linalg.norm([x, y], axis=0) <= r
    return np.array([x[mask], y[mask]])

# cdef np.ndarray range_slice(np.ndarray img, int x, int y, int length=1):
#     cdef int top, left, down, right
#     top = max(0, x-length)
#     left = max(0, y-length)
#     down = max(len(img), x+length+1)
#     right = max(len(img[0]), x+length+1)
#     return img[top:down, left:right]

cdef class Planner():

    cdef public int NEIGHBOR_RANGE
    cdef public np.ndarray NEIGHBOR_MASK, ANGLE_CRITERION
    cdef public double MOVE_COST, TURN_COST

    cdef public np.ndarray occupancy_map, local_criterion, circum_mask
    cdef public double turnable
    cdef public int path_color, obs_color, unk_color, avoidance_size, height, width, angle_num, exp_max, end_row, end_col
    cdef public vector[int] angles,
    cdef public vector[double] angle_thresh, angle_approx

    def __init__(
        self,
        np.ndarray[np.int64_t, ndim=2] occupancy_map,
        double turnable=np.pi/4,
        int path_color=0,
        int obs_color=100,
        int unk_color=-1,
        int avoidance_size=1,
        int exp_max=10000
    ):

        self.occupancy_map = occupancy_map
        self.turnable = turnable
        self.path_color = path_color
        self.obs_color = obs_color
        self.unk_color = unk_color
        self.avoidance_size = avoidance_size
        self.exp_max = exp_max

        self.NEIGHBOR_RANGE = 1
        self.NEIGHBOR_MASK = create_neighbor_mask(self.NEIGHBOR_RANGE)
        # shape: (27,3)
        self.ANGLE_CRITERION = np.arctan2(self.NEIGHBOR_MASK[:,1], self.NEIGHBOR_MASK[:,0])
        # shape: (27,)
        # [-135,  180,  135,
        #   -90,    0,   90,
        #   -45,    0,   45]

        self.height = self.occupancy_map.shape[0]
        self.width = self.occupancy_map.shape[1]
        self.end_row = self.height - self.avoidance_size
        self.end_col = self.width - self.avoidance_size
        # self.cost_map = np.empty_like(self.occupancy_map)
        self.angle_num = int(np.pi*2 // self.turnable)
        self.angles = list(range(self.angle_num))
        self.angle_thresh = [convert_180(turnable*p - turnable/2) for p in range(self.angle_num+1)]
        self.angle_approx = [convert_180(turnable*p) for p in range(self.angle_num)]
        self.local_criterion = np.array([self.get_angle(a) for a in self.ANGLE_CRITERION])
        self.circum_mask = create_circum_mask(self.avoidance_size)

        self.MOVE_COST = 2.0
        self.TURN_COST = self.MOVE_COST / self.angle_num

    cdef int get_angle(self, double rad):
        cdef int i, angle
        angle = -1
        for i in range(self.angle_num):
            if self.angle_thresh[i] < self.angle_thresh[i+1]:
                if self.angle_thresh[i] <= rad < self.angle_thresh[i+1]:
                    angle = i
                    break
            else:
                if (self.angle_thresh[i] <= rad <= np.pi) or (-np.pi <= rad < self.angle_thresh[i+1]):
                    angle = i
                    break
        return angle

    cpdef int get_angle_py(self, double rad):
        return self.get_angle(rad)

    cdef double cost_move(self, int current_x, int current_y, int current_ang, int target_x, int target_y, int target_ang):
        # cdef double cost = self.MOVE_COST
        cdef double cost = 2.0
        # if (self.occupancy_map[current_x, current_y] == self.path_color) and (self.occupancy_map[target_x, target_y] == self.unk_color):
        #     cost += 1.0
        # if self.obs_color in range_slice(self.occupancy_map, target_x, target_y, self.avoidance_size):
        #     cost += 8.0
        # if (current_x != target_x) and (current_y != target_y):
        #     cost += 0.2

        ### make like be DWB Planner ####
        # if current_ang != target_ang:
        #     cost += self.TURN_COST
        #################################
        
        # cost += 0.2 * (abs(current_ang - target_ang) % (self.angle_num-1))
        return cost

    cdef double cost_h(self, int current_x, int current_y, int current_ang, int target_x, int target_y, int target_ang):
        # return <double>max(abs(current_x-target_x), abs(current_y-target_y))
        return (abs(current_x-target_x) + abs(current_y-target_y)) / 2

    cdef np.ndarray get_neighbor(self, int x, int y, int ang):
        cdef int i, angle_diff
        cdef np.ndarray[np.int32_t, ndim=2] neighbors = np.full_like(self.NEIGHBOR_MASK, [x,y,ang], dtype=np.int32)
        cdef np.ndarray[np.uint8_t, ndim=1] result = np.full([len(neighbors)], False, dtype=np.bool8)

        cdef np.ndarray[np.int64_t, ndim=2] occupancy = self.occupancy_map
        cdef np.ndarray[np.int64_t, ndim=2] mask = self.NEIGHBOR_MASK
        cdef np.ndarray[np.int64_t, ndim=1] lc = self.local_criterion
        cdef np.ndarray[np.int64_t, ndim=2] circum = self.circum_mask

        for i in range(len(neighbors)):
            neighbors[i,0] += mask[i,0]
            neighbors[i,1] += mask[i,1]
            neighbors[i,2] = (neighbors[i,2] + mask[i,2]) % self.angle_num

            angle_diff = abs(lc[i]-ang)

            ####### filtering ##########
            if ((angle_diff <= 1) or (angle_diff >= self.angle_num-1)) and \
                ((neighbors[i,0] != x) or (neighbors[i,1] != y)) and \
                ((self.avoidance_size <= neighbors[i,0] < self.end_row) and \
                (self.avoidance_size <= neighbors[i,1] < self.end_col)):

                if not (self.obs_color in occupancy[circum[0]+neighbors[i,0], circum[1]+neighbors[i,1]]):
                    result[i] = True
        
        return neighbors[result]

    cpdef list get_path(self, tuple start_pos, tuple goal_pos):
        cdef np.ndarray[np.int32_t, ndim=4] oya
        cdef np.ndarray[np.uint8_t, ndim=3] close
        cdef list open_list
        cdef tuple goal, start
        cdef bint found
        cdef double score, score_d

        oya = np.full([self.height, self.width, self.angle_num, 3], -1, dtype=np.int32)
        close = np.full([self.height, self.width, self.angle_num], False, dtype=np.bool8)
        open_list = []
        heapq.heapify(open_list)

        goal = (round(goal_pos[0]), round(goal_pos[1]), self.get_angle(goal_pos[2]))
        start = (round(start_pos[0]), round(start_pos[1]), self.get_angle(start_pos[2]))

        if self.obs_color in self.occupancy_map[self.circum_mask[0]+goal[0], self.circum_mask[1]+goal[1]]:
            return []

        score = self.cost_h(start[0], start[1], start[2], goal[0], goal[1], goal[2])
        found = False
        heapq.heappush(open_list, (score, start))

        cdef tuple n
        cdef int i, c
        cdef np.ndarray[np.int32_t, ndim=2] neighbors

        c = 0
        while (len(open_list) != 0) and (c != self.exp_max):

            n = heapq.heappop(open_list)
            node = n[1]
            if close[node[0], node[1], node[2]]:
                continue
            if (node[0] == goal[0]) and (node[1] == goal[1]) and (node[2] == goal[2]):
                found = True
                break
            else:
                close[node[0], node[1], node[2]] = True

            neighbors = self.get_neighbor(node[0], node[1], node[2])
            # print(<list>neighbors)
            for i in range(len(neighbors)):
                if not close[neighbors[i][0], neighbors[i][1], neighbors[i][2]]:
                    score_d = score - self.cost_h(node[0], node[1], node[2], goal[0], goal[1], goal[2]) + self.cost_move(node[0], node[1], node[2], neighbors[i][0], neighbors[i][1], neighbors[i][2]) + self.cost_h(neighbors[i][0], neighbors[i][1], neighbors[i][2], goal[0], goal[1], goal[2])
                    # print(n[1], score_d, i)
                    oya[neighbors[i][0], neighbors[i][1], neighbors[i][2]] = node
                    heapq.heappush(open_list, (score_d, tuple(neighbors[i])))
                    score = min(score, score_d)
            
            c += 1

        path = deque()
        cdef tuple now

        if found:
            # print('found!')
            now = goal
            path.appendleft(goal)
            while True:
                if ((now[0] == start[0]) and (now[1] == start[1]) and (now[2] == start[2])):
                    break
                now = tuple(oya[now])
                path.appendleft(now)

        return <list>path


import os
import numpy as np

# See if Cython is installed
try:
    from Cython.Build import cythonize
# Do nothing if Cython is not available
except ImportError:
    # Got to provide this function. Otherwise, poetry will fail
    def build(setup_kwargs):
        pass
# Cython is installed. Compile
else:
    from setuptools import Extension
    from setuptools.dist import Distribution
    from distutils.command.build_ext import build_ext

    # This function will be executed in setu`p.py:
    def build(setup_kwargs):
        # The file you want to compile
        extensions = [
            "nav_sim_modules/nav_components/mapping.pyx",
            "nav_sim_modules/nav_components/planning.pyx"
        ]

        # gcc arguments hack: enable optimizations
        os.environ['CFLAGS'] = '-O3'

        # Build
        setup_kwargs.update({
            'ext_modules': cythonize(
                extensions,
                language_level=3,
                language='c++',
                compiler_directives={'linetrace': True},
            ),
            # 'extra_compile_args': ['-fopenmp'],
            # 'extra_link_args': ['-fopenmp'],
            'cmdclass': {'build_ext': build_ext},
            'include_dirs': np.get_include()
        })


from ..scener import Scener
from trimesh.path.polygons import sample
from randoor.generator import ChestSearchRoomGenerator, ChestSearchRoomConfig
import numpy as np
from shapely.geometry import Polygon

from typing import List, Tuple

from ... import SPAWN_EXTENSION, PASSABLE_COLOR, MAP_OBS_VAL, MAP_PASS_VAL, RESOLUTION, ENV_SIZE

class ChestSearchRoomScener(Scener):
    passable_color = PASSABLE_COLOR #移動可能の色　この色以外は障害物とみなす
    map_obs_val = MAP_OBS_VAL # 地図の障害物の色
    map_pass_val = MAP_PASS_VAL # 地図における移動可能

    def __init__(self, spawn_extension: float=SPAWN_EXTENSION, env_size: int=ENV_SIZE, resolution: float=RESOLUTION) -> None:

        self.tag_obstacle = 'obstacle'
        self.tag_key = 'key'
        self.tag_chest = 'chest'

        ## 直近の情報 ##
        self.room_config: ChestSearchRoomConfig = None
        self.env_pixel: np.ndarray = None
        self.sample_area: Polygon = None
        self.freespace_area: Polygon = None
        self.components_info: dict = {self.tag_obstacle: [], self.tag_key: [], self.tag_chest: []}
        ##############
        self.spawn_extension = spawn_extension
        self.env_size = env_size
        self.resolution = resolution
        self.generator_list = []
        self.parameter_list = []        

    def _generate_room(self, *args) -> ChestSearchRoomConfig: 
        if args in self.parameter_list:
            generator = self.generator_list[self.parameter_list.index(args)]
        else:
            generator = ChestSearchRoomGenerator(*args)
            self.generator_list.append(generator)
            self.parameter_list.append(args)

        return generator.generate_new()

    def spawn(self) -> Tuple[float, float, float]:
        '''
        Return the initial agent pose and truth occupancy map.
        '''
        xy = []
        while len(xy) == 0:
            xy = sample(self.sample_area, 1)
        yaw = (np.random.rand()*2-1)*np.pi
        return (xy[0][0], xy[0][1], yaw)

    def spawn_with_map(self) -> Tuple[Tuple[float,float,float], np.ndarray]:
        '''
        Return the initial agent pose.
        '''
        pose = self.spawn()
        occ_map = self.room_config.get_occupancy_grid(
            space_poly=self.freespace_area, 
            origin_pos=tuple(pose[:2]),
            origin_ori=pose[2],
            resolution=self.resolution, 
            map_size=self.env_size, 
            pass_color=self.map_pass_val, 
            obs_color=self.map_obs_val,
        ).astype(np.int64)
        return pose, occ_map.reshape([self.env_size, self.env_size]).T

    def generate_scene(self, 
                        obstacle_count=10,
                        obstacle_size=0.7,
                        target_size=0.2,
                        key_size=0.2,
                        obstacle_zone_thresh=1.5,
                        distance_key_placing=0.7,
                        range_key_placing=0.3, 
                        room_length_max=9,
                        room_wall_thickness=0.05, 
                        wall_threshold=0.1) -> None:

        self.room_config = self._generate_room(
            obstacle_count, 
            obstacle_size, 
            target_size, 
            key_size, 
            obstacle_zone_thresh,
            distance_key_placing, 
            range_key_placing,
            room_length_max, 
            room_wall_thickness, 
            wall_threshold
        )

        self.setup()

        self.components_info[self.tag_obstacle] = self.room_config.get_positions(self.room_config.tag_obstacle)
        self.components_info[self.tag_key] = self.room_config.get_positions(self.room_config.tag_key)
        self.components_info[self.tag_chest] = self.room_config.get_positions(self.room_config.tag_target)

    def setup(self) -> None:
        self.sample_area = self.room_config.get_freezone_poly().buffer(-self.spawn_extension)
        self.freespace_area = self.room_config.get_freespace_poly()
        self.env_pixel = self.room_config.get_occupancy_grid(
            space_poly=self.freespace_area, 
            resolution=self.resolution, 
            map_size=self.env_size, 
            pass_color=self.map_pass_val, 
            obs_color=self.map_obs_val
        ).astype(np.int64).reshape([self.env_size, self.env_size]).T

    def tweak_key_collision(self, index: int, is_collision: bool) -> None:
        self.room_config.tweak_key_collision(index, is_collision)

    def tweak_chest_collision(self, index: int, is_collision: bool) -> None:
        self.room_config.tweak_target_collision(index, is_collision)

    def get_env_pixel(self) -> np.ndarray:
        return self.env_pixel


import trimesh
import numpy as np

class Scener():
    def __init__(self) -> None:
        ## 直近の情報 ##
        self.room_config = None
        self.env_pixel = None
        self.components_info = {}
        ##############

    def _pixelize(self) -> np.ndarray:
        pass    

    def _generate_room(self, **kwargs):
        pass
    
    def spawn(self):
        pass


## [-12.8, 12.8]
# RESOLUTION = 0.2
# ENV_SIZE = 128
RESOLUTION = 0.1
ENV_SIZE = 256
# RESOLUTION = 0.05
# ENV_SIZE = 512

PASSABLE_COLOR = 0
MAP_OBS_VAL = 100
MAP_PASS_VAL = 0
MAP_UNK_VAL = -1

SPAWN_EXTENSION = 0.3


import numpy as np

cimport numpy as np
cimport cython

cdef class Mapper():

    cdef public np.ndarray surface, occupancy_map
    cdef public int agent_x, agent_y, size
    cdef public int passable_color, map_obs_val, map_pass_val, map_unk_val

    def __init__(
        self, 
        np.ndarray[np.int64_t, ndim=2] surface, 
        tuple agent_initial_pos, 
        int passable_color=0, 
        int map_obs_val=100, 
        int map_pass_val=0, 
        int map_unk_val=-1
    ):

        self.surface = surface
        self.size = len(self.surface)
        self.agent_x = round(agent_initial_pos[0])
        self.agent_y = round(agent_initial_pos[1])
        self.passable_color = passable_color
        self.map_obs_val = map_obs_val
        self.map_pass_val = map_pass_val
        self.map_unk_val = map_unk_val
        self.occupancy_map = np.full_like(self.surface, map_unk_val, dtype=np.int)

    @property
    def surfarray(self):
        return self.surface

    @property 
    def agent_pos(self):
        return tuple(self.agent_x, self.agent_y)

    cpdef void set_agent_pos(self, new_pos):
        self.agent_x = round(new_pos[0])
        self.agent_y = round(new_pos[1])

    cpdef void scan(self):
        cdef int i, l
        l = self.size-1
        for i in range(self.size):
            self.ray(self.agent_x, self.agent_y, 0, i)
            self.ray(self.agent_x, self.agent_y, l, i)
            self.ray(self.agent_x, self.agent_y, i, 0)
            self.ray(self.agent_x, self.agent_y, i, l)

    cdef void ray(self, int base_x, int base_y, int target_x, int target_y):
        cdef int x, y, dx, dy, sx, sy, err
        cdef np.ndarray[np.int64_t, ndim=2] buf = self.occupancy_map
        cdef np.ndarray[np.int64_t, ndim=2] surf = self.surface

        x = base_x
        y = base_y
        dx = abs(target_x - x)
        dy = abs(target_y - y)
        if x < target_x: sx = 1
        else: sx = -1
        if y < target_y: sy = 1
        else: sy = -1
        err = dx - dy

        while True:
            if (x >= self.size) or (x < 0) or (y >= self.size) or (y < 0): 
                break
            if surf[x,y] == self.passable_color:
                buf[x,y] = self.map_pass_val
            else:
                if buf[x,y] == self.map_obs_val:
                    break
                buf[x,y] = self.map_obs_val
                break
            
            e2 = 2 * err
            if e2 > -dy:
                err -= dy
                x += sx
            if e2 < dx:
                err += dx
                y += sy



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nav_sim_modules',
 'nav_sim_modules.actioner',
 'nav_sim_modules.actioner.hueristic_autonomous_actioner',
 'nav_sim_modules.nav_components',
 'nav_sim_modules.scener',
 'nav_sim_modules.scener.chest_search_room']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.4.0,<9.0.0',
 'cython>=0.29.24,<0.30.0',
 'numba>=0.54.1,<0.55.0',
 'numpy>=1.20,<2.0',
 'randoor>=0.1.4.6,<0.2.0.0']

setup_kwargs = {
    'name': 'nav-sim-modules',
    'version': '0.3.4',
    'description': 'Components for software to simulate the environment of robot that use autonomous movement system such as ROS Navigation Stack.',
    'long_description': None,
    'author': 'Reona Sato',
    'author_email': 'www.shinderu.www@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/wwwshwww/nav-sim-modules',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



from typing import Tuple
import numpy as np
from ..actioner import Actioner
from .autonomous import HueristicNavigationStack

from ... import RESOLUTION

class HeuristicLocalAutonomousActioner(Actioner):
    def __init__(
        self, 
        path_exploration_count: int=20000,
        path_planning_count: float=10, 
        allowable_angle: float=np.pi/8,
        allowable_norm: float=0.5,
        avoidance_size: int=1,
        move_limit: int=-1,
        resolution=RESOLUTION
    ) -> None:

        super().__init__(resolution)
        self.path_exploration_count = path_exploration_count
        self.allowable_angle = allowable_angle
        self.allowable_norm = allowable_norm
        self.avoidance_size = avoidance_size
        self.path_planning_count = path_planning_count
        self.move_limit = move_limit

        self.navs: HueristicNavigationStack = None
        self.occupancy_map: np.ndarray = None
    
    def initialize(self, env_pixel, global_pose) -> None:
        '''
        env_pixel should be transformed image.
        '''
        super().initialize(env_pixel, global_pose)
        self.navs = HueristicNavigationStack(
            self.env_pixel, 
            (self.local_pose_x,self.local_pose_y,self.local_pose_yaw),
            self.path_exploration_count,
            self.path_planning_count,
            self.allowable_angle,
            self.allowable_norm,
            self.avoidance_size,
            self.move_limit,
            self.resolution
        )
        self.occupancy_map = self.navs.mapper.occupancy_map

    def register_env_pixel(self, env_pixel: np.array) -> None:
        self.navs.mapper.surface = env_pixel

    def do_action(self, action) -> None:
        reached_pose = self.navs.goto(action) # should be pose of the local frame
        self.occupancy_map = self.navs.mapper.occupancy_map
        self.local_pose_x = reached_pose[0]
        self.local_pose_y = reached_pose[1]
        self.local_pose_yaw = reached_pose[2]

    def do_action_visualize(self, action, filename='move') -> None:
        reached_pose = self.navs.goto_visualize(action, filename) # should be pose of the global frame
        self.occupancy_map = self.navs.mapper.occupancy_map
        self.global_pose_x = reached_pose[0]
        self.global_pose_y = reached_pose[1]
        self.global_pose_yaw = reached_pose[2]    

    @property
    def pose(self) -> Tuple[float,float,float]:
        return (self.local_pose_x, self.local_pose_y, self.local_pose_yaw)

class HeuristicAutonomousActioner(Actioner):
    def __init__(
        self, 
        path_exploration_count: int=20000,
        path_planning_count: float=10, 
        allowable_angle: float=np.pi/8,
        allowable_norm: float=0.5,
        avoidance_size: int=1,
        move_limit: int=-1,
        resolution=RESOLUTION
    ) -> None:

        super().__init__(resolution)
        self.path_exploration_count = path_exploration_count
        self.allowable_angle = allowable_angle
        self.allowable_norm = allowable_norm
        self.avoidance_size = avoidance_size
        self.path_planning_count = path_planning_count
        self.move_limit = move_limit

        self.navs: HueristicNavigationStack = None
        self.occupancy_map: np.ndarray = None
    
    def initialize(self, env_pixel, global_pose) -> None:
        '''
        env_pixel should use image that using same frame as global_pose's.
        '''
        super().initialize(env_pixel, global_pose)
        self.navs = HueristicNavigationStack(
            self.env_pixel, 
            (self.global_pose_x,self.global_pose_y,self.global_pose_yaw),
            self.path_exploration_count,
            self.path_planning_count,
            self.allowable_angle,
            self.allowable_norm,
            self.avoidance_size,
            self.move_limit,
            self.resolution
        )
        self.occupancy_map = self.navs.mapper.occupancy_map

    def register_env_pixel(self, env_pixel: np.array) -> None:
        self.navs.mapper.surface = env_pixel

    def do_action(self, action) -> None:
        reached_pose = self.navs.goto(action) # should be pose of the global frame
        self.occupancy_map = self.navs.mapper.occupancy_map
        self.global_pose_x = reached_pose[0]
        self.global_pose_y = reached_pose[1]
        self.global_pose_yaw = reached_pose[2]

    def do_action_visualize(self, action, filename='move') -> None:
        reached_pose = self.navs.goto_visualize(action, filename) # should be pose of the global frame
        self.occupancy_map = self.navs.mapper.occupancy_map
        self.global_pose_x = reached_pose[0]
        self.global_pose_y = reached_pose[1]
        self.global_pose_yaw = reached_pose[2]

    @property
    def pose(self) -> Tuple[float,float,float]:
        return (self.global_pose_x, self.global_pose_y, self.global_pose_yaw)



from typing import Tuple
import numpy as np
from nav_sim_modules.nav_components.mapping import Mapper
from nav_sim_modules.nav_components.planning import Planner
import time
from ...utils import con2pix, pix2con

from ... import MAP_UNK_VAL, MAP_OBS_VAL, MAP_PASS_VAL, PASSABLE_COLOR, RESOLUTION
from PIL import Image

class HueristicNavigationStack():
    passable_color = PASSABLE_COLOR #移動可能の色　この色以外は障害物とみなす
    map_obs_val = MAP_OBS_VAL# 地図の障害物の色
    map_pass_val = MAP_PASS_VAL # 地図における移動可能
    map_unk_val = MAP_UNK_VAL #地図における未知の領域
    
    def __init__(
        self, 
        env_pixel: np.ndarray, 
        initial_pose: Tuple, 
        path_exploration_count: int,
        path_planning_count: int, 
        allowable_angle: float,
        allowable_norm: float,
        avoidance_size: int,
        move_limit: int=-1,
        resolution: float=RESOLUTION
    ) -> None:

        self.env_pixel = env_pixel
        self.pose = initial_pose # continuous 
        self.path_exploration_count = path_exploration_count
        self.allowable_angle = allowable_angle
        self.allowable_norm = allowable_norm
        self.avoidance_size = avoidance_size
        self.path_planning_count = path_planning_count
        self.move_limit = move_limit
        self.resolution = resolution

        self.circum_mask = create_circum_mask(self.avoidance_size)

        self.pix_center_x = len(self.env_pixel) // 2
        self.pix_center_y = len(self.env_pixel[0]) // 2

        self.mapper = Mapper(
            surface=self.env_pixel, 
            agent_initial_pos=self.con2pix(self.pose)[:2], 
            passable_color=self.passable_color, 
            map_obs_val=self.map_obs_val, 
            map_pass_val=self.map_pass_val, 
            map_unk_val=self.map_unk_val
        )
        self.mapper.scan()
        self.planner = Planner(
            occupancy_map = self.mapper.occupancy_map,
            turnable=self.allowable_angle,
            path_color=self.passable_color,
            obs_color=self.map_obs_val,
            unk_color=self.map_unk_val,
            avoidance_size= self.avoidance_size,
            exp_max=self.path_exploration_count
        )

    def con2pix(self, continuous_pos) -> tuple:
        '''
        連続値座標系をピクセル座標系に変換する関数
        ''' 
        return con2pix(continuous_pos, [self.pix_center_x,self.pix_center_y], self.resolution)

    def pix2con(self, pixel_pos) -> tuple:
        '''
        ピクセル座標系を連続値座標系に変換する関数
        ''' 
        return pix2con(pixel_pos, [self.pix_center_x, self.pix_center_y], self.resolution)

    def goto(self, goal) -> Tuple:
        pix_goal = self.con2pix(goal)
        self.mapper.set_agent_pos(self.con2pix(self.pose)[:2])
        self.mapper.scan()
        planning_count = 0
        footprint = 0
        ops_flag = False
        while planning_count < self.path_planning_count:
            if  (np.linalg.norm(np.array(goal[:2])-self.pose[:2]) <= self.allowable_norm) and\
                ((abs(self.pose[2] - goal[2]) <= self.allowable_angle) or\
                (abs(self.pose[2] - goal[2]) >= (np.pi*2 - self.allowable_angle))):
                    break
            
            path = np.array(self.planner.get_path(start_pos=self.con2pix(self.pose), goal_pos=pix_goal))
            
            if len(path) == 0:
                break
            else:
                path_mask = create_path_mask_with_mask(self.mapper.occupancy_map, path, self.circum_mask, self.map_unk_val)

                replan_flag = False
                current = self.pose
                for i in range(len(path)):
                    current = self.pix2con(path[i])
                    self.mapper.set_agent_pos(tuple(path[i][:2]))
                    self.mapper.scan()
                    footprint += 1

                    if ((footprint >= self.move_limit) and (self.move_limit != -1)) or (self.mapper.occupancy_map[pix_goal[0],pix_goal[1]] == self.map_obs_val):
                        ops_flag = True
                        break
                    if self.map_obs_val in self.mapper.occupancy_map[path_mask]:
                        replan_flag = True
                        break

                if ops_flag:
                    self.pose = (current[0], current[1], self.planner.angle_approx[current[2]])
                    break
                elif replan_flag:
                    self.pose = (current[0], current[1], self.planner.angle_approx[current[2]])
                else:
                    self.pose = (current[0], current[1], goal[2])

            planning_count +=1

        return self.pose

    def goto_visualize(self, goal, output_filename) -> Tuple:
        pixlize_start = self.con2pix((self.pose[0], self.pose[1]))
        pix_goal = self.con2pix(goal)
        pixlize_start_coler = (255, 0, 0)
        pixlize_goal_coler = (0, 0, 255)
        pixlize_path_coler = (0, 255, 0)
        pixlize_current_coler = (255, 0, 255)
        pixlize_trajectoly_coler = (255, 255, 0)
        pixlize_angle_color = (0, 255, 255)
        pixlize_angle_len = 3
        pixlize_pics = []
        pixlize_trajectoly = [pixlize_start]
        pixlize_trajectoly_mask = np.full_like(self.mapper.occupancy_map, False, dtype=np.bool8)
        pixlize_trajectoly_mask[pixlize_start[0], pixlize_start[1]] = True
        
        print(f'Navigation started: {self.con2pix(self.pose)} to {pix_goal}')
        start = time.time()
        self.mapper.set_agent_pos(self.con2pix(self.pose)[:2])
        self.mapper.scan()
        planning_count = 0
        footprint = 0
        ops_flag = False
        while planning_count < self.path_planning_count:
            if  (np.linalg.norm(np.array(goal[:2])-self.pose[:2]) <= self.allowable_norm) and\
                ((abs(self.pose[2] - goal[2]) <= self.allowable_angle) or\
                (abs(self.pose[2] - goal[2]) >= (np.pi*2 - self.allowable_angle))):
                    print('reached!')
                    break
            
            print(f'plan: {planning_count+1}')
            path = np.array(self.planner.get_path(start_pos=self.con2pix(self.pose), goal_pos=pix_goal))
            
            if len(path) == 0:
                print('path not found...')
                break
            else:
                path_mask = create_path_mask_with_mask(self.mapper.occupancy_map, path, self.circum_mask, self.map_unk_val)
                full_path_mask = np.full_like(self.mapper.occupancy_map, False, dtype=np.bool8)
                full_path_mask[path[:,0],path[:,1]] = True

                replan_flag = False
                current = self.pose
                for i in range(len(path)):
                    current = self.pix2con(path[i])
                    self.mapper.set_agent_pos(tuple(path[i][:2]))
                    self.mapper.scan()
                    footprint += 1

                    pixlize_trajectoly.append(path)
                    pixlize_trajectoly_mask[path[i][0],path[i][1]] = True

                    pixlize_pic = np.empty((*self.mapper.occupancy_map.shape,3), dtype=np.uint8)
                    pixlize_pic[:,:,0] = self.mapper.occupancy_map
                    pixlize_pic[:,:,1] = self.mapper.occupancy_map
                    pixlize_pic[:,:,2] = self.mapper.occupancy_map
                    pixlize_pic[full_path_mask] = pixlize_path_coler
                    pixlize_pic[pixlize_trajectoly_mask] = pixlize_trajectoly_coler 
                    pixlize_pic[pixlize_start[0], pixlize_start[1]] = pixlize_start_coler
                    pixlize_pic[pix_goal[0], pix_goal[1]] = pixlize_goal_coler
                    pixlize_pic[path[i][0],path[i][1]] = pixlize_current_coler
                    ang = self.planner.angle_approx[path[i][2]]
                    x = round(path[i][0]+np.cos(ang)*pixlize_angle_len)
                    y = round(path[i][1]+np.sin(ang)*pixlize_angle_len)
                    pixlize_pic[x,y] = pixlize_angle_color
                    pixlize_pics.append(pixlize_pic)

                    # print(np.sum(self.mapper.occupancy_map[path_mask]))
                    if ((footprint >= self.move_limit) and (self.move_limit != -1)) or (self.mapper.occupancy_map[pix_goal[0],pix_goal[1]] == self.map_obs_val):
                        ops_flag = True
                        print('ops...')
                        break
                    if self.map_obs_val in self.mapper.occupancy_map[path_mask]:
                        replan_flag = True
                        break                

                if ops_flag:
                    self.pose = (current[0], current[1], self.planner.angle_approx[current[2]])
                    break
                elif replan_flag:
                    self.pose = (current[0], current[1], self.planner.angle_approx[current[2]])
                else:
                    self.pose = (current[0], current[1], goal[2])

            planning_count +=1

        print(f'Time: {time.time()-start}')
        if len(pixlize_pics) > 1:
            create_gif(pixlize_pics, output_filename)    
        
        return self.pose

def create_circum_mask(r: int):
    x, y = np.meshgrid(np.arange(-r,r+1), np.arange(-r,r+1))
    mask = np.linalg.norm([x, y], axis=0) <= r
    return np.array([x[mask], y[mask]])

from numba import njit, b1, i8, prange

@njit(b1[:,:](i8[:,:], i8[:,:], i8, i8))
def create_path_mask(occupancy_map: np.ndarray, path_xy: np.ndarray, avoidance: int, map_unk_color: int):
    row = len(occupancy_map)
    col = len(occupancy_map[0])
    output_mask = np.zeros_like(occupancy_map, dtype=np.bool8)
    for i in prange(len(path_xy)):
        top = max(0, path_xy[i][0]-avoidance)
        bottom = min(row, path_xy[i][0]+avoidance+1)
        left = max(0, path_xy[i][1]-avoidance)
        right = min(col, path_xy[i][1]+avoidance+1)
        output_mask[top:bottom, left:right] = occupancy_map[top:bottom, left:right] == map_unk_color
    return output_mask

@njit(b1[:,:](i8[:,:], i8[:,:], i8[:,:], i8))
def create_path_mask_with_mask(occupancy_map: np.ndarray, path_xy: np.ndarray, mask: np.ndarray, map_unk_color: int):
    output_mask = np.zeros_like(occupancy_map, dtype=np.bool8)
    l = len(mask[0])
    for i in prange(len(path_xy)):
        x = mask[0]+path_xy[i][0]
        y = mask[1]+path_xy[i][1]
        for j in prange(l):
            if occupancy_map[x[j]][y[j]] == map_unk_color:
                output_mask[x[j]][y[j]] = 1
    return output_mask


def create_gif(frames: list, filename: str="output"):
    frs = [Image.fromarray(f, mode="RGB") for f in frames]
    frs[0].save(f'{filename}.gif', save_all=True, append_images=frs[1:], optimize=False, duration=60, loop=0)


from abc import abstractmethod
from typing import Tuple, List

from ..utils import con2pix, pix2con
from .. import RESOLUTION

class Actioner():
    def __init__(self, resolution=RESOLUTION) -> None:
        self.action_space = None
        self.observation_space = None
        self.env_pixel = None
        self.resolution = RESOLUTION

        self.pix_center_x = 0
        self.pix_center_y = 0

        self.global_pose_x = 0
        self.global_pose_y = 0
        self.global_pose_yaw = 0

        self.local_pose_x = 0
        self.local_pose_y = 0
        self.local_pose_yaw = 0
        
    def initialize(self, env_pixel, global_pose: Tuple) -> None:
        self.env_pixel = env_pixel
        self.pix_center_x = len(self.env_pixel) // 2
        self.pix_center_y = len(self.env_pixel[0]) // 2
        self.global_pose_x = global_pose[0]
        self.global_pose_y = global_pose[1]
        self.global_pose_yaw = global_pose[2]

    @abstractmethod
    def do_action(self, action: any) -> None:
        pass

    @property
    def pose(self):
        return None

    def con2pix(self, pose):
        return con2pix(pose, [self.pix_center_x,self.pix_center_y],self.resolution)

    def pix2con(self, pose):
        return pix2con(pose, [self.pix_center_x,self.pix_center_y],self.resolution)



from typing import Tuple, Sequence

import numpy as np
from decimal import Decimal, ROUND_HALF_UP

from . import PASSABLE_COLOR, RESOLUTION

def rgb2int(rgb: Tuple[int,int,int]) -> int:
    '''
    converting RGB to int: 
    '(255,255,255)' -> '0xffffff' -> '16777215'
    '''
    return (rgb[0] << (4*4)) + (rgb[1] << (4*2)) + (rgb[2])

def int2rgb(num: int) -> Tuple[int,int,int]:
    '''
    converting int to RGB: 
    '16777215' -> '0xffffff' -> '(255,255,255)'
    '''
    r, g, b = hex(num)[2:4], hex(num)[4:6], hex(num)[6:8]
    return (int(r, 16), int(g, 16), int(b, 16))

def half_up(target_float: float) -> int:
    d = Decimal(str(target_float))
    return int(d.quantize(Decimal('0'), rounding=ROUND_HALF_UP))

def convert_180(rad360):
    '''
    [0,360] -> [-180,180]
    '''
    return ((rad360 - np.pi) % (np.pi*2)) - np.pi

def convert_360(rad180):
    '''
    [-180,180] -> [0,360]
    '''
    return rad180 % (np.pi*2)

def con2pix(continuous_pos: Sequence, pix_center, resolution=RESOLUTION) -> tuple:
    '''
    Convert continuous pose to pixel pose
    ''' 
    pos_1 = round(continuous_pos[0] / resolution + pix_center[0])
    pos_2 = round(continuous_pos[1] / resolution + pix_center[1])

    return (pos_1, pos_2, *continuous_pos[2:])

def pix2con(pixel_pos: Sequence, pix_center, resolution=RESOLUTION) -> tuple:
    '''
    Convert pixel pose to continuous pose
    ''' 
    pos_1 = (pixel_pos[0] - pix_center[0]) * resolution
    pos_2 = (pixel_pos[1] - pix_center[1]) * resolution

    return (pos_1, pos_2, *pixel_pos[2:])